'Wav sound player for Quick Media Player
'Copyright (C) 2013 Raymond Allen.
'  MIT License:   See end of file for terms of use.
'Ver2:  Adding graphical UI
'RevA:  Accounted for metadata before "fmt " and "data"
'Notes:

CON
  _CLKMODE      = XTAL1 + PLL16X                       
  _XINFREQ      = 5_000_000

CON 'SD card pin settings
  SD_MISO=27
  SD_Clk=26
  SD_MOSI=25
  SD_CS=24
  
CON 'Audio Pins and buffer size
  'Here you can select the audio output pins
  'Quick Media Player Settings
  Pin_Right=23
  Pin_Left=22
  
CON 'Buffer and volume settings
  'Slow SD cards may require a larger buffer, adjust with value below
  buffLongs=100
  buffWords = buffLongs*2+1 
  'Volume control
  volume = 35535  'range is 0 to 65535 with 65535 being the loudest
  MaxPlayList = 100 'maximum number of wav files in root to add to playlist
  

OBJ
  lcd : "QMP_DisplayDriver2d"  'Controls the LCD
  sd: "fsrw"     'SD card driver
  ser : "fullduplexserial" 'for monitoring
  i2c: "QMP_I2CDriver"  


VAR  'Display variables
  long foregroundcolor,backgroundcolor
  
VAR 'buffer for bitmap line of pixels
  long lineBuffer[320*3/4]' buffer for line of an image and scrolling  'needs to be long aligned
  long BmpPalette[256]
  
VAR 'Assembly wav player variables
    long parameter1  'to pass @buff1 to ASM
    long parameter2  'to pass @buff2 to ASM
    long parameter3  'to pass sample rate to ASM
    long parameter4  'to pass #samples to ASM
    long parameter5  'to pass #channels to ASM
    long parameter6  'to pass #samples in each buffer
    long parameter7  'to pass volume to ASM
    word buff1[buffWords]
    word buff2[buffWords]
    byte Header[44]

VAR 'touchscreen
  long bStillDown, touchx, touchy
  
DAT 'define on-screen buttons
nButtons byte 1
Buttons byte
buttonPlay byte 152,193,173,216,1  'x1,x2,y1,y2,handler 


DAT 'Default Touchscreen Calibrations
'based on these reading:
'  Top of landscape screen: x=60600, Bottom of landscape screen: x=4560   
'  Left of landscape screen (screen only, not buttons): y=6200,  Right of landscape screen:  y=60600
'  Left of button area:  (y=1500)  
touchx0 long 4560
touchdx long (60600-4560)/240
touchy0 long 6200
touchdy long (60600-6200)/320
    
pub Main|i,j               'Startup Routine
  'Start serial connection
  ser.start(31,30,0,9600)
  'initialize the LCD
  lcd.start
  'init touch
  i2c.StartTouch
  
  'set default colors
  backgroundcolor:=lcd.RGB(0,199,140) 'Light green
  foregroundcolor:=lcd.RGB(255,255,0) 'yellow
  'Fill screen fast way
  lcd.FillScreen(backgroundcolor)
  lcd.SetFontColors(foregroundcolor,backgroundcolor)   
  'Change font
  lcd.SetFont(@Font1)  

  lcd.DrawString(100,10,string("QMP Wav Player"))

  'Mounting SD card now
  ser.str(string("Attempting to mount SD card",13))
  repeat   'try forever to mount the SD card
    i:=\sd.Mount_Explicit(SD_MISO,SD_CLK,SD_MOSI,SD_CS) 
    if i==0
       quit  'mounted, so exit
    waitcnt(cnt+clkfreq/4)
    ser.tx(".")

  bmp(50,240-51-20,string("PlayBG.bmp"))  'Play button with light green background
  'lcd.DrawRectangle(lcd#rgb_white,buttonPlay[0],buttonPlay[2],buttonPlay[1],buttonPlay[3])  'see where button is
  

  CreatePlaylist
  ser.tx(13)
  ser.dec(nPlayList)
  ser.str(string(" wav files found."))
  ser.tx(13)

  repeat        'wait until play button pushed
    WaitStylusDown
    WaitStylusUp
    Ser.tx("*")
    ser.dec(touchx)
    ser.tx(" ")
    ser.dec(touchy)
  until GetButton==1
     bmp(50,240-51-20,string("PlayBP.bmp"))'Pressed Play button glows green
  'lcd.DrawRectangle(lcd#rgb_white,buttonPlay[0],buttonPlay[2],buttonPlay[1],buttonPlay[3]) 

  repeat   'loop playlist
    repeat i from 1 to nPlayList
       j:=@Playlist+(i-1)*13  'filename address
       ser.str(string(13,"Playing: "))
        
        lcd.DrawString(10,40,string("STRAY Saturday Morning Pictures"))
       ser.str(j)
       ser.tx(13)
       'show filename
       lcd.DrawSolidRectangle(backgroundcolor,50,70,50+9*8,70+16)  'clear old text
       lcd.DrawString(50,70,j)
       'play
       PlayWav(j) 


PUB GetButton|i,p 'see if a button was pushed
  p:=@Buttons
  repeat i from 0 to nButtons-1
    if touchx>byte[p] and touchx<byte[p+1]
      if touchy>byte[p+2] and touchy<byte[p+3]
        return byte[p+4]
    p+=5
  
  return 0  'no button found
  

PUB bTouchDown|tx,ty,z1,z2,tz  
  tx:=i2c.GetTouchX
  z1:=i2c.GetTouchZ1
  z2:=i2c.GetTouchZ2
  tz:=tx*(z2-z1)/z1
  if (tz<90000) and (tz>10000)           'You might have to tweak these values for some displays...
    return true
  else
    return false

PUB WaitStylusUp
  'Wait for stylus to be up
  waitcnt(cnt+clkfreq/10)
  repeat  
    if !bTouchDown
    return

    
PUB WaitStylusDown|tx,ty,z1,z2,tz      'Note: x&y reversed in landscape mode 
  'Wait for stylus to be down then update touchx, touchy and return
  repeat
    if bTouchDown
      tx:=i2c.GetTouchX
      ty:=i2c.GetTouchY
      z1:=i2c.GetTouchZ1
      z2:=i2c.GetTouchZ2
      'take another x,y reading and average
      tx+=i2c.GetTouchX
      ty+=i2c.GetTouchY
      tx>>=1
      ty>>=1
      tz:=tx*(z2-z1)/z1
      if (tz<90000) and (tz>10000)
        touchy:=240-(tx-touchx0)/touchdx      'Note: x&y reversed in landscape mode
        touchx:=(ty-touchy0)/touchdy
        return
      else
        bStillDown:=false
    else
      bStillDown:=false   

PUB CreatePlaylist|i,j
  sd.opendir
  repeat  'get next file
    i:=sd.nextfile(@fname)
    if i<0
      quit  'no more files
    if IsWavFilename
      ser.tx(13)
      nPlayList++
      j:=@Playlist+(nPlayList-1)*13 'destination
      bytemove(j,@fname,12)
      ser.str(j)  

PUB IsWavFilename:bWav|i,j  'see if filename is a .wav file
  i:=@fname
  repeat
    if byte[i]=="."
      quit
    i++
  j:=byte[++i]
  if j<>"W" and j<>"w"
    return false
  j:=byte[++i]
  if j<>"A" and j<>"a"
    return false
  j:=byte[++i]
  if j<>"V" and j<>"v"
    return false
  return true
   

PUB PlayWav(pFileName)|n,i,j, SampleRate,Samples,nChannels,ChunkSize,SamplePeriod,PlayTime
  'open file
  ser.str(string("Attempting to open wav file:  "))
  ser.str(pFileName)
  ser.tx(13)
                   
  i:=\sd.popen(pFileName,"r")             
  if (i<>0)
    ser.str(string("Cannot open wav file.",13))  
    return

  'Find fmt
  ser.str(string("Looking for fmt..."))
  i:=sd.pread(@Header, 12)
  repeat
    sd.pread(@Header[12], 1)
    if Header[12]=="f"
      sd.pread(@Header[13], 1)
      if Header[13]=="m"
        sd.pread(@Header[14], 1)
        if Header[14]=="t"        
          sd.pread(@Header[15], 1)
          if Header[15]==" "
            quit  'found fmt
            
  'Plays only   16-bit PCM WAV Files !!!!!!!!!!!!!!!!
  'See here for header format:  https://ccrma.stanford.edu/courses/422/projects/WaveFormat/ 
  i:=sd.pread(@Header+16, 44-8-16) 'read data words to input stereo buffer
  'Get bits per sample (must be 16 (or maybe 32))
  if (Header[35]<<8+Header[34])<>16
    ser.str(string(13,"Error:  Not 16 bits per sample!",13))
    ser.dec(  Header[35]<<8+Header[34])
    'return 'Note:  Microsoft Media Player seems to rip with this set as 32 bits per sample... 
  'Get audio format, must be PCM
  if (Header[21]<<8+Header[20])<>1
    ser.str(string(13,"Error:  Not uncompressed PCM data",13))
    return       
  'Get sample rate from header
  SampleRate:=Header[27]<<24+Header[26]<<16+Header[25]<<8+Header[24]
  ser.str(string("Sample Rate: ")) 
  ser.dec(SampleRate)
  ser.tx(13)
  'Get # channels from header
  nChannels:=Header[23]<<8+Header[22]
  ser.str(string("#Channels: "))
  ser.dec(nChannels)
  ser.tx(13)
  'Find data
  ser.str(string("Looking for data...")) 
  repeat
    sd.pread(@Header[36], 1)
    if Header[36]=="d"
      sd.pread(@Header[37], 1)
      if Header[37]=="a"
        sd.pread(@Header[38], 1)
        if Header[38]=="t"        
          sd.pread(@Header[39], 1)
          if Header[39]=="a"
            quit  'found data
            
  sd.pread(@Header[40],4) 'read subchunk size    
  'get # samples from header
  ChunkSize:=Header[43]<<24+Header[42]<<16+Header[41]<<8+Header[40]
  Samples:=(ChunkSize>>1)/nChannels
  ser.str(string("#Samples: ")) 
  ser.dec(Samples)
  ser.tx(13)
  'Calculate play time
  SamplePeriod:=CLKFREQ/SampleRate  '#clocks between samples'1814'for 44100ksps,  5000 'for 16ksps 
  ser.str(string("Playtime (s) = ")) 
  ser.dec(Samples/SampleRate)
  ser.tx(13)


    

  'Start ASM player in a new cog
  ser.str(STRING("Running ASM Player",13))    
  parameter1:=@buff1                   'address of first buffer
  parameter2:=@buff2                   'address of second buffer
  parameter3:=SamplePeriod             'time between samples
  parameter4:=Samples                  'total # samples
  parameter5:=nChannels                '#channels
  parameter6:=(BuffWords-1)/nChannels  '#samples in each buffer
  parameter7:=volume&$FFFFF            'volume control (0..65535)
  COGNEW(@ASMWAV,@parameter1)

  'Keep filling buffers until end of file
  ' note:  using alternating buffers to keep data always at the ready...
  n:=buffWords-1
  buff1[n]:=1  'set empty flags at end of buffers
  buff2[n]:=1
  i:=n*2   'number of bytes to read

  'Keep buffers filled until got all samples
  j:=samples*nChannels*2
  PlayTime :=cnt
  repeat
    repeat until buff1[n]==1 'wait until buffer empty
    if sd.pread(@buff1, i)<>i  'read data words to input stereo buffer       
      quit 
    buff1[n]:=0 'clear empty flag
    j-=i
    if j=<0
      quit
    
    repeat until buff2[n]==1  'wait until buffer empty
    if sd.pread(@buff2, i)<>i
      quit 'read data words to input stereo buffer   
    buff2[n]:=0 'clear empty flag
    j-=i
    if j=<0
      quit

  ser.str(string("Played time (s) = ")) 
  ser.dec((cnt-PlayTime)/clkfreq)
  ser.tx(13)   

  'must have reached the end of the file, so close it
  \sd.pclose
  
  return

PUB BMP(bmpX,bmpY,sFilename)|i,j,k,l
'Draw a bitmap file onto the screen
    'Update speed seems to be only limited by SD card reading speed
    'note that bitmaps are stored from bottom line to top line!
    i:=\sd.popen(sFilename,"r") 'open image for reading
    if i<>0
      return i   
    sd.pread(@bfType,2) 'read bmp header
    sd.pread(@bfSize,4) 'read bmp header  
    sd.pread(@bfReserved1,4) 'read bmp header  
    sd.pread(@bfOffBits,16) 'read bmp header
    sd.pread(@biPlanes,4)
    sd.pread(@biCompression,24)

    biWidth:=214

    if (biHeight<1) or (biWidth<1)
      return -1  'some kind of problem

    l:=biHeight+bmpY-1  
      
    case biBitCount
      24:  'true color (3 bytes per pixel)
        i:=(biWidth*3)//4
        k:=biWidth*3+i
        repeat j from l to bmpY 
          'process image row by row          
          sd.pRead(@lineBuffer,k)

          lcd.SetWindow(bmpX,constant(lcd#xpixels-1),j,constant(lcd#ypixels-1)) 
          lcd.WriteBitmapLine(@lineBuffer,biWidth,24,0)
 
      8:  '256-Palette mode
        'first need to read in palette  (rgbquad (4 bytes) * # pallete entries)
        'calculate actual bytes in palette
        i:=bfOffBits-54
        'read in palette
        sd.pread(@BmpPalette,i)
        i:=biWidth//4
        k:=biWidth+i
        repeat j from l to bmpY          
          sd.pRead(@lineBuffer,k)
          lcd.SetWindow(bmpX,constant(lcd#xpixels-1),j,constant(lcd#ypixels-1)) 
          lcd.WriteBitmapLine(@lineBuffer,biWidth,8,@BmpPalette) 
      
      4:  '16-Palette mode
        'first need to read in palette  (rgbquad (4 bytes) * # pallete entries)
        'calculate actual bytes in palette
        i:=bfOffBits-54        
        'read in palette
        sd.pread(@BmpPalette,i)
        'now paint picture
        i:=(biWidth>>1)//4
        k:=(biWidth>>1)+i 
        repeat j from l to bmpY          
          sd.pRead(@lineBuffer,k)
          lcd.SetWindow(bmpX,constant(lcd#xpixels-1),j,constant(lcd#ypixels-1)) 
          lcd.WriteBitmapLine(@lineBuffer,biWidth,4,@BmpPalette) 

    sd.pclose
    return 0
 
DAT
  ORG 0
ASMWAV
'load input parameters from hub to cog given address in par
        movd    :par,#pData1             
        mov     x,par
        mov     y,#7  'input 7 parameters
:par    rdlong  0,x
        add     :par,dlsb
        add     x,#4
        djnz    y,#:par

setup
        'setup output pins
        MOV DMaskR,#1
        ROL DMaskR,OPinR
        OR DIRA, DMaskR
        MOV DMaskL,#1
        ROL DMaskL,OPinL
        OR DIRA, DMaskL
        'setup counters
        OR CountModeR,OPinR
        MOV CTRA,CountModeR
        OR CountModeL,OPinL
        MOV CTRB,CountModeL
        'Wait for SPIN to fill table
        MOV WaitCount, CNT
        ADD WaitCount,BigWait
        WAITCNT WaitCount,#0
        'setup loop table
        MOV LoopCount,SizeBuff
        MOV pData,pData1
        MOV nTable,#1
        'setup loop counter
        MOV WaitCount, CNT
        ADD WaitCount,dRate



MainLoop
        CMP nSamples,#0 wz,wc 'see if out of samples
        IF_e JMP #Done
        SUB nSamples,#1
        waitcnt WaitCount,dRate
        cmp   nChannel,#2 wz,wc
        if_e jmp #Stereo     'otherwise, assume mono
Mono
        call  #ReadWithVolume
        mov   right,y
        mov   left,right
        jmp #Output
        
Stereo
        call  #ReadWithVolume
        mov   right,y
        add pData,#2
        call  #ReadWithVolume
        mov   left,y 

Output
        ADD pData,#2       
        MOV FRQA,Right        
        MOV FRQB,Left
        

        'loop
        DJNZ LoopCount,#MainLoop

        WRword one,pData  'set empty flag 
        
        MOV LoopCount,SizeBuff        
        'switch table       ?
        CMP nTable,#1 wz
        IF_Z JMP #SwitchToTable2
SwitchToTable1
        MOV nTable,#1
        MOV pData,pData1
        JMP #MainLoop
SwitchToTable2
        MOV nTable,#2
        MOV pData,pData2
        JMP #MainLoop
        
                
Done
        
         'now stop
        COGID thisCog
        COGSTOP thisCog          

'Multiply routine adapted from manual
'' Multiply x[15..0] by y[15..0] (y[31..16] must be 0)
' on exit, product in y[31..0]
ReadWithVolume
        mov   x,dVolume
        rdword y,pData
        add   y,twos
        and   y,ffff
        shl x,#16 'get multiplicand into x[31..16]
        mov t,#16 'ready for 8 multiplier bits
        shr y,#1 wc 'get initial multiplier bit into c
:loop
        if_c add y,x wc 'if c set, add multiplicand to product
        rcr y,#1 wc 'put next multiplier in c, shift prod.
        djnz t,#:loop 'loop until done

ReadWithVolume_ret
        ret 'return with product in y[31..0]


'Working variables
thisCog long 0
x       long 0
y       long 0
t       long 0
dlsb    long    1 << 9
BigWait long 100000
twos    long $8000_8000
ffff    long $ffff
        
'Loop parameters
nTable  long 0
WaitCount long 0
pData   long 0
LoopCount long 0
Left    long 0
Right   long 0
Zero    long 0
one     long 1       

'setup parameters
DMaskR  long 0 'right output mask
OPinR   long Pin_Right 'right channel output pin #                   
DMaskL  long 0 'left output mask 
OPinL   long Pin_Left 'left channel output pin #                       
CountModeR long %00011000_00000000_00000000_00000000
CountModeL long %00011000_00000000_00000000_00000000


'input parameters
pData1   res 1  'Address of first data table        
pData2   res 1 'Address of second data table
dRate    res 1 'clocks between samples
nSamples res 1 '#samples to play
nChannel res 1 '#channels (1 or 2)
SizeBuff res 1 '#samples in each buffer
dVolume  res 1 'volume byte



DAT     'Data for reading bitmap files

BMPHeader  byte 'Mostly using info from here:  http://www.fortunecity.com/skyscraper/windows/364/bmpffrmt.html
bfType byte "B","M"  ' 19778
bfSize long 0
bfReserved1 word 0
bfReserved2 word 0
bfOffBits long 54
biSize long 40
biWidth long 0
biHeight long 10
biPlanes word 1
biBitCount word 24
biCompression long 0
biSizeImage long 0
biXPelsPerMeter long 0
biYPelsPerMeter long 0
biClrUsed long 0
biClrImportant long 0    


DAT  'Font:  Custom font located here.  Other fonts can be selected with the SetFont command
'Note:  If you do not need a font, you can save space by commenting out the file command line below
        
font1 long
file "ProgoCo_wHeader.dat"   'nice 8x16 fixed pitch font

DAT   'for storing filename from SD card
fname   byte "i"[13],0

DAT 'playlist
nPlayList   long 0
PlayList      byte 0 [8*MaxPlayList]

DAT 'MIT License
{{
'  Permission is hereby granted, free of charge, to any person obtaining
'  a copy of this software and associated documentation files
'  (the "Software"), to deal in the Software without restriction,
'  including without limitation the rights to use, copy, modify, merge,
'  publish, distribute, sublicense, and/or sell copies of the Software,
'  and to permit persons to whom the Software is furnished to do so,
'  subject to the following conditions:
'
'  The above copyright notice and this permission notice shall be included
'  in all copies or substantial portions of the Software.
'
'  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
'  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
'  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
'  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
'  CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
'  TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
'  SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
}}